import numpy as np
from collections import deque
from phylokrr.treeio import get_vcv

def one_hot(cat_feat):
    tmp_uniq = list(set(cat_feat))
    # print(tmp_uniq)
    feat_encod = []
    for tmp_str in cat_feat:
        tmp_encod = [0.]*len(tmp_uniq)
        tmp_encod[tmp_uniq.index(tmp_str)] = 1.0
        feat_encod.append(tmp_encod) 
        
    return feat_encod

def get_Xy(data_file, verbose = False, delimiter = ',', y_col = -1):
    """
    Read data file and return predictors and response variable.
    
    Parameters
    ----------
    data_file : str
        path to the data file

    verbose : bool
        print information about the data

    delimiter : str
        delimiter of the data file

    y_col : int
        column index of the response variable

    Returns
    -------
    X : numpy array
        predictors

    y : numpy array
        response variable
    """

    data = np.loadtxt(data_file, 
                         delimiter=delimiter, 
                         dtype=str)
    
    assert data.shape[0] > 0, "data file must have at least one row"

    try:
        y = data[:,y_col].astype(float)

    except ValueError:
        assert False, "y_col must be a numeric column"

    data = np.delete(data, y_col, axis = 1)
    
    n,p = data.shape
    assert p > 0, "predictors must have at least one column"

    k = 0 
    col_struct = deque([])
    # check if columns are numeric or categorical
    # and store the future column indices
    for j in range(p):
        try:
            data[0,j].astype(float)
            col_struct += [k]
            k += 1

        except ValueError:
            col_span = len(set(data[:,j]))
            col_struct += [ list(range(k, k + col_span)) ]
            k += col_span

    # col_struct
    X = np.zeros((n, k))

    float_cols = deque([])
    cat_cols = deque([])
    j = 0

    for col in col_struct:
        if isinstance(col, int):
            X[:,col] = data[:,j].astype(float)
            float_cols += [col]

        else:
            X[:,col] = one_hot(data[:,j])
            cat_cols += [col]

        j += 1

    # verbose = True
    if verbose:
        print("X shape: ", X.shape)
        print("numeric columns: ", list(float_cols))
        print("categorical columns: ", list(cat_cols))

    return X, y

def linTime_intersection(names_vcv, spps_ff, verbose = False):
    
    # O(n)
    vcv_dict = {names_vcv[i] : i for i in range(len(names_vcv))}

    vcv_ord = deque([])
    dat_ord = deque([])
    # O(n)
    for i in range(len(spps_ff)):
        tmp_spps  = spps_ff[i]

        # all data file rows must 
        # contain correlation information
        if not tmp_spps in vcv_dict:
            if verbose:
                print(f"{tmp_spps} not found in the tree file")
            continue

        vcv_ord.append( vcv_dict[tmp_spps] )
        dat_ord.append( i )

    return list(vcv_ord), list(dat_ord)


def intersect_w_spps(names_vcv, spps_file, X, y, delimiter = ',', verbose = False):
    """
    intersect species names from the data file and the tree file

    Parameters
    ----------
    names_vcv : numpy array
        species names from the tree file

    spps_file : str
        path to the species names file

    X : numpy array
        predictors

    y : numpy array
        response variable

    delimiter : str
        delimiter of the species names file

    verbose : bool
        print information about the species names

    Returns
    -------
    vcv_ord : list
        indices of the species names in the tree file
    
    dat_ord : list
        indices of the species names in the data file
    """

    # species names from file
    spps_ff = np.loadtxt(spps_file, delimiter=delimiter, dtype=str)

    assert len(spps_ff) == len(set(spps_ff)), "species names must be unique"
    assert len(spps_ff) == X.shape[0], "the number of species names must match the number of rows in the data file"


    # match spps_ff and names
    vcv_ord, dat_ord = linTime_intersection(names_vcv, spps_ff, verbose = verbose)
    assert len(vcv_ord) > 0, "species names must match the tree and data file"

    # verbose = True
    if verbose:
        print(f"{len(spps_ff)} names from data, {len(names_vcv)} from tree, {len(vcv_ord)} intersect")

    return vcv_ord, dat_ord

def read_data(tree_file, data_file, data_file_spps, y_col = -1, delimiter = ',', verbose = False):
    """
    read data file and tree file.
    If there are categorical variables in X, they are one-hot encoded.
    If the response variable is not numeric, an assertion error is raised.
    To see where the categorical variables are, set verbose to True.


    Parameters
    ----------
    tree_file : str
        path to the tree file

    data_file : str
        path to the data file. This data file
        does not contain column names or row names

    data_file_spps : str
        path to the species names file. This data file
        contains only the species names. This is used to order
        the data_file based on the species names in the rows 
        of the tree_file's VCV matrix.

    y_col : int
        column index of the response variable

    delimiter : str
        delimiter of the data file

    verbose : bool
        print information about the data

    Returns
    -------
    X : numpy array
        predictors
    
    y : numpy array
        response variable

    vcv : numpy array
        variance-covariance matrix. This matrix is ordered
        based on the species names in the data file.
        The item i,j in the matrix represents the correlation 
        between species i and j and it is obtained by calculating 
        the distance between the root and the lowest common
        ancestor of species i and j in the tree.
    """

    with open(tree_file, 'r') as f:
        tree = f.read().strip()

    vcv, names = get_vcv(tree)

    names_vcv = np.array(names)
    vcv = np.array(vcv)

    X, y = get_Xy(data_file,
                  verbose = verbose, 
                  y_col=y_col, 
                  delimiter=delimiter)
    
    
    vcv_ord, dat_ord = intersect_w_spps(names_vcv, data_file_spps, X, y, 
                                 delimiter=delimiter,
                                 verbose = verbose)
    
    vcv = vcv[vcv_ord,:][:,vcv_ord]

    X = X[dat_ord,:]
    y = y[dat_ord]

    return X, y, vcv


# from phylokrr.utils import weight_data

# tree_file      = "./data/test_tree.txt"
# data_file_spps = "./data/test_data_spps.txt"
# data_file      = "./data/test_data3.csv"

# X_uw_uc0, y_uw_uc0, vcv0 = read_data(tree_file, data_file, data_file_spps,
#                                  y_col = 1, delimiter = ',', verbose = True)

# X_w0, y_w0 = weight_data(X_uw_uc0, y_uw_uc0, vcv0, use_sd = False)

# np.random.seed(12038)

# import matplotlib.pyplot as plt

# fig, axs = plt.subplots(1, 2, figsize=(8, 4))
# axs[0].scatter(X_uw_uc0, y_uw_uc0, color = 'red', alpha=0.5, )
# axs[0].set_title('Unweighted data')
# axs[0].set_xlabel('x')
# axs[0].set_ylabel('y')
# axs[1].scatter(X_w0, y_w0, color = 'blue', alpha=0.5, )
# axs[1].set_title('Phylogenetically weighted data')
# axs[1].set_xlabel('$x^*$')
# axs[1].set_ylabel('$y^*$')
# plt.tight_layout()


# Internal original file location: get_Xy_cov.py from LS code base

# with open(tree_file, 'r') as f:
#     tree = f.read().strip()
#     vcv, names = get_vcv(tree)  

# vcv = np.array(vcv)
# with open(data_file_spps, 'w') as f:
#     f.write('\n'.join(names))


# np.random.seed(12037)
# mean_vector = np.zeros(vcv.shape[0])

# X_w = np.random.normal(0, 1, vcv.shape[0])
# # Non-linear response variable (sine curve)
# y_w = np.sin(X_w*1.5).ravel() 

# # Add noise to the response variable
# y_w[::10] += 4 * (0.5 - np.random.rand(X_w.shape[0] // 10))

# # we can attempt to unweight to the original space
# # with the square root of the covariance matrix
# P_inv = P_inv_simple(vcv)
# X_uw_uc, y_uw_uc = P_inv @ X_w, P_inv @ y_w

# import matplotlib.pyplot as plt

# fig, axs = plt.subplots(1, 2, figsize=(8, 4))
# axs[0].scatter(X_uw_uc, y_uw_uc, color = 'red', alpha=0.5, )
# axs[0].set_title('Unweighted data')
# axs[0].set_xlabel('x')
# axs[0].set_ylabel('y')
# axs[1].scatter(X_w, y_w, color = 'blue', alpha=0.5, )
# axs[1].set_title('Phylogenetically weighted data')
# axs[1].set_xlabel('$x^*$')
# axs[1].set_ylabel('$y^*$')
# plt.tight_layout()



# # Xy = np.stack([X_uw_uc, y_uw_uc], axis=1)
# # np.savetxt('../src/data/test_data3.csv', Xy, delimiter=',')






import random
import numpy as np

# seed for reproducibility
seed = 12038 
np.random.seed(seed)
random.seed(seed)



import numpy as np
from phylokrr.metrics import rmse, R2

def distance_matrix(a, b):
    """
    l2 norm squared matrix
    """
    return np.linalg.norm(a[:, None, :] - b[None, :, :], axis=-1)**2

def RBF_kernel(a, b, gamma):
    """
    Radial Basis Function
    """        
    tmp_rbf = -gamma * distance_matrix(a, b)
    np.exp(tmp_rbf, tmp_rbf) # RBF kernel. Inplace exponentiation
    return tmp_rbf

def linear_kernel(a, b, c):
    """
    Linear Kernel
    """
    XXt = a.dot(b.T)
    C = c * np.ones(XXt.shape)

    return XXt + C



class KRR:

    def __init__(self, kernel = 'rbf', 
                 fit_intercept = True, 
                 check_cov = False,
                 store_gram = False
                 ) -> None:
        """
        Weighted Kernel Ridge Regression

        kernel: str, default = 'rbf'
            kernel type

        fit_intercept: bool, default = True
            fit intercept term

        check_cov: bool, default = False
            Check covariance matrix, i.e., P.T @ P == vcv^-1

        store_gram: bool, default = False
            store the gram matrix used to fit the model
        """

        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        self.fit_intercept = fit_intercept
        self.check_cov = check_cov
        self.store_gram = store_gram
        
        # internal
        self.intercept = 0
        self.alpha = np.array([])
        self.X = np.array([])
        self.K = np.array([])

        self.chol = False

    def set_params(self, **params):

        if self.kernel == 'rbf':
            self.params['gamma'] = params['gamma']

        else:
            self.params['c'] = params['c']

        self.params['lambda'] = params['lambda']

    def get_params(self):
        return self.params
    
    def P_mat(self, vcv):
        """
        get the square root of the inverse of the
        covariance matrix.

        vcv: np.array, default = None
            covariance matrix
        """
        
        if isinstance(vcv, type(None)):
            return None
        
        if self.check_cov:
            self.assert_COV_sym(vcv)

        if self.chol:
            C = np.linalg.cholesky( vcv )
            P = np.linalg.inv( C )

        else:
            L,Q  = np.linalg.eig( vcv )
            P  = Q @ np.diag( L**(-1/2) ) @ Q.T 

        if self.check_cov:
            self.assert_COV_decom(P, vcv)

        return P
    
    def assert_COV_sym(self, vcv, tol=1e-8):
        assert np.all(np.abs(vcv-vcv.T) < tol), 'not symmetric matrix'

    def assert_COV_decom(self, P, vcv):
        """
        check if P.T @ P == vcv^-1
        expensive calculation for large matrices
        """
        assert np.all(np.round(P.T @ P, 2) == np.round(np.linalg.inv(vcv), 2)), "P.T @ P != vcv^-1"

    def Gram_mat(self, X1, X2):
        """
        Gram matrix

        X1: np.array
            data matrix (e.g., testing data)

        X2: np.array
            data matrix (e.g., training data)
        """

        if self.kernel == 'rbf':
            K = RBF_kernel(X1, X2, self.params['gamma'])

        else:
            K = linear_kernel(X1, X2, self.params['c'])

        return K
        

    def fit(self, X, y, vcv = None):
        """
        Fit the model

        vcv: np.array, default = None
            covariance matrix

        """
        # X = X_train
        # y = y_train
        # vcv = vcv_train
        
        self.X = X
        # self.y = y

        P = self.P_mat(vcv)
        K_train = self.Gram_mat(self.X, self.X)

        if self.store_gram:
            self.K = K_train

        self.alpha = self.opt_alpha(K_train, y, self.params['lambda'], P)
        
        # intercept term 
        if self.fit_intercept:
            self.intercept = np.mean(y - K_train @ self.alpha)

    def predict(self, X_test):

        assert len(self.alpha) > 0, "The model needs to be fitted first"

        K_test = self.Gram_mat(X_test, self.X)
        
        return K_test @ self.alpha + self.intercept
    
    def score(self, X_test, y_test, vcv_test, metric = 'rmse'):
        """
        
        Calculates score considerint the covariance matrix
        This error is the same as the in objective function

        X_test: np.array
            testing data
        
        y_test: np.array
            testing targets
        
        vcv_test: np.array
            testing covariance matrix. If None, then it
            will return the standard RMSE and R^2

        metric: str, default = 'rmse'
            weighted root mean squared error (the same 
            that was used to construct the objective function).
            if metric = 'r2', then it returns the R^2 which
            also considers the covariance matrix, see Willett & Singer (1988)

        Returns:
        --------
        float
            score

        Reference:
        ---------
        Willett, J. B., & Singer, J. D. (1988). 
        Another cautionary note about R 2: Its use 
        in weighted least-squares regression analysis. 
        The American Statistician, 42(3), 236-238.
        """

        y_pred = self.predict(X_test)
        r = y_pred - y_test
        
        if isinstance(vcv_test, type(None)):
            P = np.eye(X_test.shape[0])

        else:
            P = self.P_mat(vcv_test)

        werr = P @ r # weighted residuals
        Py   = P @ y_test # weighted targets

        if metric == 'rmse':
            return rmse(werr)
        
        else:
            return R2(Py, werr)
            
        # if metric == 'rmse':
        #     return np.sqrt( np.mean( r**2 ) )
        
        # else:
        #     u = r.T @ r
        #     v = y_test.T @ y_test - n * np.mean(y_test)**2
        #     return 1 - (u/v)
    
    def opt_alpha(self, K, y, reg_lam = None, P = None):
        # Y = y
        # X = X
        # m = None
        n,_ = self.X.shape
        
        I = np.eye(K.shape[0])
        nlI = n * reg_lam * I

        if isinstance(P, type(None)):
            return np.linalg.solve(K + nlI, y)
        
        else:
            return P @ np.linalg.solve(P @ K @ P + nlI,  P @ y)
            # return np.linalg.solve(P.T @ P @ K + nlI,  P.T @ P @ y)


# cov. matrix obtained from the phylogenetic tree
vcv = np.loadtxt("../data/test_cov2.csv", delimiter=',') 

# Trait simulation under Brownian motion
n = vcv.shape[0]
mean = np.zeros(n)
X = np.random.multivariate_normal(cov=vcv, mean=mean).reshape(-1,1)
# Non-linear response variable (sine curve)
y = np.sin(X*2).ravel() + 5 

# Add noise to the response variable
y[::10] += 4 * (0.5 - np.random.rand(X.shape[0] // 10)) 

from phylokrr.utils import split_data_vcv

# split data into training and testing sets 
num_test = round(0.5*n)

(X_train  , X_test,  
 y_train  , y_test,  
 vcv_train, vcv_test) = split_data_vcv(X, y, vcv, num_test, seed = seed) # seed defined above



# set model
model = KRR(kernel='rbf', fit_intercept= True)

# arbitrarily proposed hyperparameters
params = {'lambda': 2, 'gamma': 2}

# set hyperparamters
model.set_params(**params)

import time
start = time.time()
# fit model with phylogenetic covariance matrix
model.fit(X_train, y_train, vcv = vcv_train)
y_pred1 = model.predict(X_test)
rmse1 = model.score(X_test, y_test, vcv_test, metric='rmse')
r1 = model.score(X_test, y_test, vcv_test, metric='r2')

model.fit(X_train, y_train, vcv =None)
y_pred2 = model.predict(X_test)
rmse2 = model.score(X_test, y_test, None, metric='rmse')
r2 = model.score(X_test, y_test, None, metric='r2')
end = time.time()
print("Time: ", end - start)


# import matplotlib.pyplot as plt
# from phylokrr.utils import PGLS

# # fit standard phylogenetic regression
# b_wls = PGLS(X_train, y_train, vcv_train)
# y_pred3 = np.hstack((np.ones((X_test.shape[0],1)), X_test)) @ b_wls

# plt.scatter(X_test, y_test , color = 'blue' , alpha=0.5, label = 'Testing (unseen) data')
# plt.scatter(X_test, y_pred1, color = 'green', alpha=0.5, label = 'phyloKRR predictions w\o CV')
# plt.scatter(X_test, y_pred3, color = 'red', alpha=0.5, label = 'PGLS predictions')
# plt.xlabel('x')
# plt.ylabel('y')
# plt.legend()

from scipy.optimize import minimize
import numpy as np

# )
model = KRR(kernel='rbf',
            fit_intercept = True,
            store_gram = True)
model.fit(X, y, vcv = vcv)

K = model.K
a = model.alpha
l = model.params['lambda']


def objective_function(K, y, l, vcv, lambda_val):
    n,p = X.shape
    # lambda_val = 0.05

    W = np.diag( np.diag(vcv) )
    C = lambda_val * vcv + (1 - lambda_val) * W
    
    nlI = n * l * np.eye(n)
    OpK = np.linalg.solve(C, K)
    Opy = np.linalg.solve(C, y)

    a = np.linalg.solve(OpK+ nlI, Opy)
    e = y - K @ a
    
    dual = (1/n) * e.T @ np.linalg.solve(C, e) + l*( a.T @ K @ a )



    

    C_inv = np.linalg.inv(C)
    beta  = np.linalg.inv(X.T @ C_inv @ X) @ (X.T @ C_inv @ y)
    e = y - X @ beta 
    s = (1/n)*e.T @ C_inv @ e

    objective = n*np.log(s) + np.log(np.linalg.det(C)) + (1/s)*e.T@ C_inv @e
    return objective.astype(np.float64)[0][0]

def optimize_problem(X, y, Sigma):
    n, p = X.shape
    X = np.hstack((np.ones((n, 1)), X))

    # Define the bounds for lambda
    bounds = [(0, 1)]

    # Initial guess for lambda 
    initial_lambda = np.random.uniform(0, 1)

    # Define the optimization problem
    fnc = lambda lambda_val: objective_function(X, y, Sigma, lambda_val)
    result = minimize(fnc, x0 = initial_lambda, bounds=bounds, method='SLSQP')

    return result


